library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use work.const.all;


entity ex_stage is
port (
	clk, reset, stall, nop: in std_logic;
	id_op1, id_op2, imm: in dat_type;
	opcode: in std_logic_vector(3 downto 0);
	o_ex_data: out dat_type;

	i_pc, i_id_branchAdr: in adr_type;
	branch_cond: in std_logic_vector(1 downto 0);
	i_sel8, i_sel16, i_sel32: in std_logic;
	i_isValid, i_ismem, i_isldi: in std_logic;
	i_isalu, i_isaluspec, i_ishigh, i_islow, i_isstore, i_isbranch: in std_logic;
	i_ishalt: in std_logic;
	rd, rs1, rs2: in reg_adr_type;
	i_exc_vector: in exc_vector_type;
	i_itlb_ena: in std_logic;

	o_branch_cond: out std_logic_vector(1 downto 0);
	o_sel8, o_sel16, o_sel32: out std_logic;
	o_isValid, o_ismem, o_isldi: out std_logic;
	o_isalu, o_isaluspec, o_isstore, o_isbranch, o_ishalt: out std_logic;
	o_reg_addr: out reg_adr_type;
	o_reg2: out dat_type;
	o_take_branch: out std_logic;
	o_isbranch_if: out std_logic; --goes directly to IF (not latched)
	o_pc, o_branchAdr: out adr_type;
	o_regWrEna: out std_logic;
	o_rs2Adr: out reg_adr_type;
	o_itlb_dat, o_itlb_adr: out dat_type;
	o_itlb_ena: out std_logic;

	--forwarding stuff
	ex_rdAdr: in reg_adr_type;
	ex_rd: in dat_type;
	ex_Ena: in std_logic;
	wb_rdAdr: in reg_adr_type;  -- actually, this is from mem
	wb_rd: in dat_type;			-- but we use calculated values from wb
	wb_Ena: in std_logic;

	--branch prediction
	o_branch_misspredict: out std_logic;

	o_exc_vector: out exc_vector_type
);
end entity ex_stage;

architecture behav of ex_stage is
	signal z, c: std_logic;
	signal ldi_data, alu_data, i_ex_data, ex_op2: dat_type;
	signal imm16: std_logic_vector(15 downto 0);
	signal i_reg_addr: reg_adr_type;
	signal op1, op2: dat_type;
	signal i_regWrEna: std_logic;
	signal branchAdr: adr_type;
	signal i_take_branch: std_logic;
	signal exc_overflow: std_logic;
	signal exc_vector: exc_vector_type;
begin
	--bypass
	op1 <= 	ex_rd when rs1 = ex_rdAdr and ex_Ena = '1' else
	        wb_rd when rs1 = wb_rdAdr and wb_Ena = '1' else
	        id_op1;
	op2 <=	ex_rd when rs2 = ex_rdAdr and ex_Ena = '1' else
	        wb_rd when rs2 = wb_rdAdr and wb_Ena = '1' else
	        id_op2;
	--end of bypass
	ex_op2 <= imm when i_ismem = '1' else
	          op2;

	ALU: entity work.ex_alu(behav)
		PORT MAP(opcode,
			op1, ex_op2,
			alu_data, branch_cond, i_take_branch);

	i_reg_addr <= rs2 when i_ismem = '1' or i_isldi = '1' else
	              rd when i_isalu = '1' or i_isaluspec = '1' else (others => '0');

	imm16 <= imm(15 downto 0);
	ldi_data <= op2(31 downto 16) & imm16 when i_islow = '1' else
	            imm16 & op2(15 downto 0) when i_isHigh = '1' else
	            imm;

	i_ex_data <= ldi_data when i_isldi = '1' else alu_data;

	i_regWrEna <= i_isalu or i_isaluspec or (i_ismem and not i_isstore) or i_isldi;

	o_branchAdr <= i_id_branchAdr;
	o_take_branch <= i_take_branch;
	o_isbranch_if <= i_isbranch;

	--right now, we have static branch prediction (always-not-taken)
	--so miss is when branch is taken
	--this equation is easy to change once we add smarter branch prediction
	o_branch_misspredict <= i_take_branch and i_isbranch;

	--write to iTLB
	o_itlb_dat <= op1;
	o_itlb_adr <= op2;
	o_itlb_ena <= i_itlb_ena;

	exc_overflow <= '0';
	exc_inj_overflow: entity work.exc_inject(behav)
		GENERIC MAP(exc_ex_overflow)
		PORT MAP(
			exc_overflow,
			i_exc_vector,
			exc_vector
		);

	LATCH: entity work.ex_latch(behav)
		PORT MAP(clk, reset, stall, nop,
			i_ex_data, o_ex_data,

			i_pc,
			branch_cond,
			i_sel8, i_sel16, i_sel32,
			i_isValid, i_ismem, i_isldi,
			i_isalu, i_isaluspec, i_isstore, i_isbranch, i_ishalt,
			i_reg_addr, op2,
			rs2,
			i_regWrEna,
			exc_vector,

			o_pc,
			o_branch_cond,
			o_sel8, o_sel16, o_sel32,
			o_isValid, o_ismem, o_isldi,
			o_isalu, o_isaluspec, o_isstore, o_isbranch, o_ishalt,
			o_reg_addr, o_reg2,
			o_rs2Adr,
			o_regWrEna,
			o_exc_vector
		);

end architecture behav;
