
#include <errno.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "as.h"

extern struct label labels[MAXLABELS], *curlabel;
extern struct code codes[MAXCODES], *curcode;

static unsigned long org_addr;
static volatile unsigned linenum;
static volatile struct label *linelabel;

static void
error(char *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	error_(linenum, fmt, va);
	va_end(va);
}

struct label *
lookup(char *s)
{
	struct label *lp;
	unsigned offset = codes - curcode;
	size_t len;

	for (lp = labels; lp < curlabel; ++lp) {
		if (!strcmp(lp->s, s))
			return lp;
	}
	if (curlabel == &labels[MAXLABELS])
		die("too much labels");

	len = strlen(s);
	if ((curlabel->s = malloc(len * sizeof(char))) == NULL)
		die("out of memory");
	memcpy(curlabel->s, s, len + 1);

	return curlabel++;
}

static struct label *
label(void)
{
	static char labelstr[MAXLABEL_SIZ];
	struct label *lp;
	int c;

	if (!isalpha(c = getchar()))
		error("incorrect label");
	ungetc(c, stdin);
	scanf("%8[a-zA-Z0-9]", labelstr);
	if (isalpha(c = getchar()))
		error("maximum label length is 8 characters");
	if (c != ':')
		error("label must end with a ':'");
	lp = lookup(labelstr);
	if (lp->def)
		error("redifinig label '%s'", labelstr);
	lp->def = 1;
	lp->offset = ((curcode - codes) << 2) + org_addr;
	return lp;
}

static void
newcode(struct code *code)
{
	if (curcode == &codes[MAXCODES])
		die("too much code");
	*curcode = *code;
	curcode->line = linenum;
	++curcode;
}

static void
comment(void)
{
	int c;

	while (isspace(c = getchar())) {
		if (c == '\n')
			return;
	}
	if (c != ';')
		error("garbage at the end of line");
	while ((c = getchar()) != '\n' && c != EOF)
		/* nothing */;
}

static void
fmtnopars(char op)
{
	struct code code;

	memset(&code, 0, sizeof(code));
	code.op = op;
	newcode(&code);
}

static unsigned long
immvalue(char *imm)
{
	unsigned long v;
	int len, base;
	char *end;

	if (isalpha(imm[0])) {
		struct label *lp;

		if (!(lp = lookup(imm)) || !lp->def)
			error("unknown value '%s'", imm);
		return lp->offset;
	}

	len = strlen(imm);
	base = (toupper(imm[len - 1]) == 'H') ? 16 : 10;
	v = strtol(imm, &end, base);
	if (base == 10 && *end ||
	    base == 16 && end != imm + len - 1) {
		error("incorrect inmediate");
	}
	return v;
}

static void
fmt(int relative, char op, char *imm,
    unsigned r1, unsigned r2, unsigned r3,
    int cond)
{
	int offset, last;
	struct code code;
	char *end;

	code.imm_islabel = 0;
	if (imm) {
		if (isalpha(imm[0])) {
			code.imm_islabel = 1;
			code.imm = lookup(imm) - labels;
		} else {
			code.imm = immvalue(imm);
		}
	}

	if (r1 > 31 || r2 > 31 || r3 > 31)
		error("incorrect register number");
	code.op = op;
	code.src1 = r1;
	code.src2 = r2;
	code.dst = r3;
	code.cond = cond;
	code.relative = relative;
	newcode(&code);
}

static void
fmt1imm2regs(char op)
{
	char buf[MAXLABEL_SIZ];
	int r1, r2, n;

	if (scanf(" %10[-+A-Za-z0-9](R%d),R%d", buf, &r1, &r2) != 3)
		error("syntax error");
	fmt(0, op, buf, r1, r2, 0, 0);
}

static void
fmt2regs1imm(char op)
{
	char buf[MAXLABEL_SIZ];
	int r1, r2;

	if (scanf(" R%d,%10[-+A-Za-z0-9](R%d)", &r1, buf, &r2) != 3)
		error("syntax error");
	fmt(0, op, buf, r1, r2, 0, 0);
}

static void
fmt1imm1reg(char op)
{
	char buf[MAXLABEL_SIZ];
	int r1;

	if (scanf(" %10[-+A-Za-z0-9],R%d", buf, &r1) != 2)
		error("syntax error");
	fmt(0, op, buf, r1, 0, 0, 0);
}

static void
fmt3regs(char op)
{
	int r1, r2, r3;

	if (scanf(" R%d,R%d,R%d", &r1, &r2, &r3) != 3)
		error("syntax error");
	fmt(0, op, NULL, r1, r2, r3, 0);
}

static void
fmt1imm(char op)
{
	char buf[MAXLABEL_SIZ];

	if (scanf(" %10[-+A-Za-z0-9]", buf) != 1)
		error("syntax error");
	fmt(0, op, buf, 0, 0, 0, 0);
}

static void
org(char op)
{
	char imm[MAXLABEL_SIZ];
	extern unsigned long org_addr;

	if (scanf(" %10[-+A-Za-z0-9]", imm) != 1)
		error("syntax error");
	org_addr = immvalue(imm);
}

static void
data(char op)
{
	char buf[MAXLABEL_SIZ];
	int c;

	do
		fmt1imm(op);
	while ((c = getchar()) == ',');

	ungetc(c, stdin);
}

static void
equ(char op)
{
	char imm[MAXLABEL_SIZ];
	int c, base, len;
	char *end;

	if (linelabel == NULL)
		error("equ must be preceded by a label");
	if ((c = getchar()) != '$') {
		ungetc(c, stdin);
		if (scanf(" %10[-+A-Za-z0-9]", imm) != 1)
			error("syntax error");
		linelabel->offset = immvalue(imm);
	}
}

static void
fill(char op)
{
	char imm1[MAXLABEL_SIZ], imm2[MAXLABEL_SIZ];
	unsigned  cnt;
	struct label *lp;

	if (scanf(" %10[-+A-Za-z0-9],%10[-+A-Za-z0-9]", imm1, imm2) != 2)
		error("syntax error");
	for (cnt = immvalue(imm1); cnt--; fmt(0, op, imm2, 0, 0, 0, 0));
		/* nothing */
}

static void
fmt2regs(char op)
{
	int r1, r2;

	if (scanf(" R%d,R%d", &r1, &r2) != 2)
		error("syntax error");
	fmt(0, op, NULL, r1, 0, r2, 0);
}

static void
jmp(char op)
{
	char buf[MAXLABEL_SIZ];

	if (scanf(" %10[-+A-Za-z0-9]", buf) != 1)
		error("syntax error");
	fmt(0, JMP, buf, 0, 0, 0, 0);
}

static void
brn(char op)
{
	char buf[MAXLABEL_SIZ],cond[3];
	int r1, r2, c, n;

	n = scanf(" %2[CZNczn],%10[-+A-Za-z0-9],R%d,R%d", cond, buf, &r1, &r2);
	if (n != 4)
		error("syntax error");
	cond[0] = toupper(cond[0]);
	cond[1] = toupper(cond[1]);
	if (!strcmp(cond, "Z"))
		c = 0;
	else if (!strcmp(cond, "NZ"))
		c = 1;
	else if (!strcmp(cond, "C"))
		c = 2;
	else if (!strcmp(cond, "NC"))
		c = 3;
	else
		error("incorrect branch condition");

	fmt(1, BRN, buf, r1, r2, 0, c);
}

static void
mov(char op)
{
	int c, reg, dst;
	char line[BUFSIZ], buf[3];

	if (!fgets(line, BUFSIZ, stdin))
		return;

	if (!strchr(line, '\n') && !feof(stdin))
		error("line too long");
	ungetc('\n', stdin);

	reg = dst = 0;
	if (sscanf(line, " R%d,R%d", &reg, &dst) == 2) {
		op = MOV;
	} else if (sscanf(line, " FA,R%d", &dst) == 1) {
		op = FMOV;
	} else if (sscanf(line, " R%d,%2[A-Za-z0-9]", &reg, buf) == 2) {
		if (strcmp(buf, "I"))
			goto syntax;
		op = IMOV;
	} else {
		goto syntax;
	}

	fmt(0, op, NULL, reg, 0, dst, 0);
	return;

syntax:	error("syntax error");
}

void
pass1(void)
{
	static struct instruction {
		char *s;
		void (*fun)(char);
		char op;
	} *bp, ins[] = {
		"NOP", fmtnopars, NOP,
		"BS", fmtnopars, BS,
		"IRET", fmtnopars, IRET,
		"HALT", fmtnopars, HALT,
		"LDB", fmt1imm2regs, LDB,
		"LDW", fmt1imm2regs, LDW,
		"LD", fmt1imm2regs, LD,
		"STB", fmt2regs1imm, STB,
		"STW", fmt2regs1imm, STW,
		"ST", fmt2regs1imm, ST,
		"LDIH", fmt1imm1reg, LDIH,
		"LDIL", fmt1imm1reg, LDIL,
		"LDI", fmt1imm1reg, LDI,
		"ADD", fmt3regs, ADD,
		"NOT", fmt2regs, NOT,
		"ITLB", fmt2regs, ITLB,
		"DTLB", fmt2regs, DTLB,
		"NEG", fmt2regs, NEG,
		"MOV", mov, MOV,
		"SUB", fmt3regs, SUB,
		"OR", fmt3regs, OR,
		"AND", fmt3regs, AND,
		"SLL", fmt3regs, SLL,
		"SRL", fmt3regs, SRL,
		"XOR", fmt3regs, XOR,
		"JMP", jmp, JMP,
		"BRN", brn, BRN,
		"DATA", data, DATA,
		"EQU", equ, EQU,
		"FILL", fill, FILL,
		"ORG", org, ORG,
		NULL, NULL, 0
	};

	for (linenum = 1; ; ++linenum) {
		char op[10];
		int c;
		extern jmp_buf recovery;

		if (setjmp(recovery)) {
			while ((c = getchar()) != '\n' && c != EOF)
				/* nothing */;
			continue;
		}

		linelabel = NULL;
		switch (c = getchar()) {
		case EOF:
			return;
		case '\v':
		case '\f':
		case '\r':
		case '\t':
		case ' ':
			while (isspace(c = getchar()) && c != '\n')
				/* nothing */;
			if (c == '\n')
				continue;
			if (c != ';') {
				ungetc(c, stdin);
				break;
			}
		case ';':
			ungetc(c, stdin);
			comment();
			continue;
		case '\n':
			continue;
		default:
			ungetc(c, stdin);
			linelabel = label();
			break;
		}

		if (scanf(" %5[A-Za-z]", op) != 1 && !feof(stdin))
			error("incorrect instruction");
		for (bp = ins; bp->s && strcmp(bp->s, op); ++bp)
			/* nothing */;
		if (!bp->s)
			error("unkown opcode '%s'", op);
		else
			(*bp->fun)(bp->op);
		comment();
	}
}

